package othello.model;

import java.rmi.* ;

import othello.observation.* ;

/**
 * <p>Title: Othello</p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Memorial University</p>
 * @author Theodore S. Norvell
 * @version 1.0
 */

public class RemoteGameModel
    extends RemotelyConcurrentlyObservable
    implements RemoteGameModelInterface
{
    private static final long serialVersionUID = -2331193617098867606L;

	/** @todo Simplify this class */
    private int [][] square = new int[ROWS][COLS] ;
    {
        for( int r=0 ; r<ROWS ; ++r ) {
            for( int c=0 ; c<COLS ; ++c ) {
                square[r][c] = EMPTY ; } }
    }
    
    private int[][] directions = new int[][]{
            {-1, -1},
            {-1,  0},
            {-1,  +1},
            { 0, -1},
            {0,  +1},
            {+1, -1},
            {+1,  0},
            {+1, +1} } ;
    
    private int state = UNINITIALIZED ;
    private int numberOfPlayers = 0 ;
    
    
    public RemoteGameModel() throws RemoteException { super() ; }
    
    public int getGameState() {
        readerEnters() ; try { return state ; } finally { readerExits() ; } }
    
    public int getColour() {
    	writerEnters() ;
    	try {
    		if( numberOfPlayers == 0) {
    			numberOfPlayers++ ;
    			return BLACK ; }
    		else if( numberOfPlayers == 1) {
    			numberOfPlayers++ ;
    			// Initialize on a new thread so that the call to getColor
    			// can complete. This is important if that call comes from
    			// a GUI thread.
    			initialize() ;
    			return WHITE ; }
    		else {
    			return EMPTY ; } }
    	finally { writerExits() ; }
    }
    
    public void initialize() {
    	if( numberOfPlayers == 2 ) {
    		System.err.println("Initializing");
    		layPiece( ROWS/2-1, COLS/2-1, WHITE ) ;
    		layPiece( ROWS/2-1, COLS/2,   BLACK ) ;
    		layPiece( ROWS/2,   COLS/2-1, BLACK ) ;
    		layPiece( ROWS/2,   COLS/2,   WHITE ) ;
    		state = BLACK_TO_MOVE ; } 
    }
    
    public int move(int row, int col, int colour ) {
    	writerEnters() ;
    	try { 
    		if( colour == BLACK && state != BLACK_TO_MOVE ) 
    			return NOT_YOUR_TURN ;
    		if( colour == WHITE && state != WHITE_TO_MOVE ) 
    			return NOT_YOUR_TURN ;        

    		if( legalMove( row, col, colour ) ) {
    			makeMove( row, col, colour ) ;
    			if( hasLegalMove( -colour ) ) {
    				state = -colour == BLACK ? BLACK_TO_MOVE : WHITE_TO_MOVE ;
    				return DONE ; }
    			else if( hasLegalMove( colour ) ) {
    				state = colour == BLACK ? BLACK_TO_MOVE : WHITE_TO_MOVE ;
    				return DONE ; }
    			else {
    				state = determineWinner() ;
    				return DONE ; } }
    		else {
    			return ILLEGAL_MOVE ; } }
    	finally{ writerExits() ; } }
        
    public int getPieceAt( int row, int col ) throws RemoteException {
        readerEnters() ; try { return square[row][col] ; } finally { readerExits() ; } }
    
    private void layPiece( int row, int col, int bw ) {
        square[row][col] = bw ;
        notifyObservers( new Pair(row,col) ); }
        
    private void flipPiece( int row, int col ) {
        if( square[row][col] == BLACK ) square[row][col] = WHITE ;
        else                            square[row][col] = BLACK ;
        notifyObservers( new Pair(row,col) ); }
    
    private boolean legalMove( int row, int col, int colour, int d ) {
        int dr = directions[d][0] ;
        int dc = directions[d][1] ;
        int r = row+dr ;
        int c = col+dc ;
        if( inBounds( r, c ) && square[r][c] == -colour ) {
            do {
                r = r+dr ;
                c = c+dc ;
                if( !inBounds( r, c ) || square[r][c] == EMPTY )
                    break ;
                if( square[r][c] == colour ) 
                    return true ;
            } while( true ) ; }
        return false ;
    }
    
    private boolean legalMove( int row, int col, int colour ) {
    	if( ! inBounds( row, col ) ) return false ;
    	else if( square[row][col] != EMPTY ) return false ;
        else {
            for( int d = 0 ; d<8 ; ++d ) {
                 if( legalMove( row, col, colour, d) ) return true ; }
            return false ; }
    }
    
    private void makeMove( int row, int col, int colour ) {
        layPiece(row, col, colour );
        for( int d = 0 ; d<8 ; ++d ) {
            if( legalMove( row, col, colour, d ) ) {
                int dr = directions[d][0] ;
                int dc = directions[d][1] ;
                int r = row+dr ;
                int c = col+dc ;
                while( inBounds( r, c ) && square[r][c] == -colour ) {
                    flipPiece(r, c);
                    r = r+dr ;
                    c = c+dc ; } } }
    }
    
    private boolean hasLegalMove( int colour ) {
        for( int row = 0 ; row < ROWS ; ++row ) {
            for( int col = 0 ; col < COLS ; ++ col ) {
                if( legalMove( row, col, colour ) ) return true ; } }
        return false ; }
    
    private int determineWinner() {
        int blackCount = 0 ;
        int whiteCount = 0 ;
        for( int row = 0 ; row < ROWS ; ++row ) {
            for( int col = 0 ; col < COLS ; ++ col ) {
                if( square[row][col] == BLACK ) blackCount++ ;
                else if( square[row][col] == WHITE ) whiteCount++ ; } }
        if( blackCount > whiteCount ) {
            return WHITE_HAS_WON ; }
        else if( whiteCount > blackCount ) {
            return BLACK_HAS_WON ; }
        else {
            return TIE_GAME ; }
    }
    
    private boolean inBounds( int row, int col ) {
        return 0 <= row && row < ROWS && 0 <= col && col < COLS ; }
}
