package othello.view;

import java.awt.* ;
/** A simple perspective transform.
 * <p>Title: Othello</p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Memorial University</p>
 * @author Theodore Norvell
 * @version 1.0
 */

public class PerspectiveViewStrategy implements ViewStrategy {

    Transform3D xform ;
    
    public PerspectiveViewStrategy() {
//  WARNING.
//  The following code is preliminary and
// is included only to illustrate the idea that
// one can do interesting things with the strategy pattern
// and in particular by varying the the model to view
// mapping while keeping the same paint routine.
// Nevertheless, preliminary or not, you have to admit
// this is kind of cool.
// END OF WARNING.  
/** @todo Add a VUP vector */
        double[] camera = new double[]{-30.0,0.0, 15.0} ;
        double[] lookAt = new double[]{5.0, 5.0, 0.0} ;
        // Rotate around the Z axis till camera is out of x-y plane
        Transform3D rotateZ = Transform3D.createRotationZ(
                Math.atan2( camera[1], camera[0] ) ) ;
        double d = Math.sqrt( camera[0]*camera[0] + camera[1]*camera[0]) ;
        // Rotate around Y axis
        Transform3D rotateY = Transform3D.createRotationY(
                Math.atan2( d, camera[2] )) ;
        Transform3D xlateUp = Transform3D.createTranslation(0, 0, camera[2]) ;
        Transform3D xlateToOrigin = Transform3D.createTranslation(
                    -lookAt[0], -lookAt[1], -lookAt[2]) ;
        xform = Transform3D.compose(
                    xlateUp,
                    Transform3D.compose(
                        rotateZ,
                        Transform3D.compose(
                            rotateY,
                             xlateToOrigin) ) ) ;
        Transform3D perspective = Transform3D.createPerspective( 20, 0.0, 0.0 ) ;
        xform = Transform3D.compose( perspective, xform ) ;
    }
    
    public Point worldToView( Component component, double row, double col ) {
        // (row, col) is a position in the world.
        // We need to (a) map it to coordinates relative to the camera
        // and (b) do a perspective projection onto the view plane.
        // Both these transformations are done by applying xform.
            double[] p = new double[]{ row, col, 0, 1.0 } ;
            p = xform.apply( p ) ;
            
            row = p[0] ;
            col = p[1] ;
        
        // Now (row, col) is a position on the view plane, where
        // the origin of the view plane is on the camera to lookat point.
        // We take the view area to be a 20.0 by 20.0 square.
        // It remains to map positions on the view plane to pixel coordinates (x,y).
            int height = component.getHeight() ;
            double vertStretch = height / 20.0 ;
            int width = component.getWidth() ;
            double horStretch = width / 20.0 ;
            horStretch = vertStretch = Math.min( horStretch, vertStretch ) ;
            int y = (int)( (10.0+row)*vertStretch ) ;
            int x = (int)( (10.0+col)*horStretch ) ;
        return new Point( x, y ) ; }
}

class Transform3D {
    double[][] a = new double [4][4] ;
    
    private Transform3D() {}
    
    static Transform3D createScale(double x, double y, double z) {
        Transform3D t = new Transform3D() ;
        t.a[0][0] = x ;
        t.a[1][1] = y ;
        t.a[2][2] = z ;
        t.a[3][3] = 1.0 ;
        return t ; }
    
    static Transform3D createTranslation(double x, double y, double z) {
        Transform3D t = new Transform3D() ;
        t.a[0][3] = x ;
        t.a[1][3] = y ;
        t.a[2][3] = z ;
        t.a[0][0] = t.a[1][1] = t.a[2][2] = t.a[3][3] = 1.0 ;
        return t ; }
    
    static Transform3D createRotationX(double theta) {
        double s = Math.sin(theta) ;
        double c = Math.cos(theta) ;
        Transform3D t = new Transform3D() ;
        t.a[0][0] = t.a[3][3] = 1.0 ;
        t.a[1][1] = t.a[2][2] = c ;
        t.a[1][2] = -s ;
        t.a[2][1] = s ;
        return t ; }
    
    static Transform3D createRotationY(double theta) {
        double s = Math.sin(theta) ;
        double c = Math.cos(theta) ;
        Transform3D t = new Transform3D() ;
        t.a[1][1] = t.a[3][3] = 1.0 ;
        t.a[0][0] = t.a[2][2] = c ;
        t.a[2][0] = -s ;
        t.a[0][2] = s ;
        return t ; }
    
    static Transform3D createRotationZ(double theta) {
        double s = Math.sin(theta) ;
        double c = Math.cos(theta) ;
        Transform3D t = new Transform3D() ;
        t.a[2][2] = t.a[3][3] = 1.0 ;
        t.a[0][0] = t.a[1][1] = c ;
        t.a[0][1] = -s ;
        t.a[1][0] = s ;
        return t ; }
        
    
    static Transform3D createPerspective(double N, double a, double b) {
        Transform3D t = new Transform3D() ;
        t.a[0][0] = N ;
        t.a[1][1] = N ;
        t.a[2][2] = a ;
        t.a[2][3] = b ;
        t.a[3][2] = -1.0 ; // Should be -1
        return t ; }
    
    static Transform3D compose( Transform3D t, Transform3D u ) {
        
        Transform3D v = new Transform3D() ;
        for( int i=0 ; i < 4 ; ++i ) {
            for( int j=0 ; j < 4 ; ++j ) {
                double s = 0.0 ;
                for( int k = 0 ; k < 4 ; ++k ) {
                    s += t.a[i][k] * u.a[k][j] ; }
                v.a[i][j] = s ; } }
        return v ; }
    
    double[]  apply( double[] p ) {
        double[] r = new double[4] ;
        for( int i=0 ; i < 4 ; ++i ) {
            double s = 0.0 ;
            for( int k = 0 ; k < 4 ; ++k ) {
                s += a[i][k] * p[k] ; }
            r[i] = s ; }
        // Normalize
        for( int i = 0 ; i < 4 ; ++i ) {
            r[i] = r[i] / r[3] ; }
        return r ; }
}